#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
from ssht00ls.classes import utils
from ssht00ls.classes.smartcards import smartcards

# the ssh key object class.
class Keys(Traceback):
	def __init__(self,
		# initialize as specific not global (optional).
		#	the username.
		username=None,
		# the path to the public key.
		public_key=None,
		# the path to the private key.
		private_key=None,
	):

		# defaults.
		Traceback.__init__(self, traceback="ssht00ls.keys", raw_traceback="ssht00ls.classes.keys.Keys")

		# specific variables.
		self.specific = username != None
		self.username = username
		self.private_key = private_key
		self.pubic_key = public_key

		#
	def edit_passphrase(self, path=None, old=None, new=None):

		# check specific.
		if self.specific:
			if path == None: path = self.private_key

		# checks.
		path = Formats.denitialize(path)
		old = Formats.denitialize(old)
		new = Formats.denitialize(new)
		response = dev0s.response.parameters.check({
			"old":old,
			"new":new,
			"path":path,
		}, traceback=self.__traceback__(function="edit_passphrase"))
		if not response["success"]: return response
		
		# check dir.
		if os.path.isdir(path): path += '/private_key'

		# do.
		output = utils.__execute__(["ssh-keygen", "-p", "-P", old, "-N", new, "-f", path])

		# check fails.
		if "incorrect passphrase supplied" in output:
			return dev0s.response.error(f"Provided an incorrect passphrase for key [{path}].")
		elif "No such file or directory" in output:
			return dev0s.response.error(f"Key [{path}] does not exist.")
		
		# check success.	
		elif "Your identification has been saved with the new passphrase" in output:
			return dev0s.response.success(f"Successfully edited the passphrase of key [{path}].")

		# unknown.
		else:
			l = f"Failed to edit the passphrase of key [{path}]"
			return dev0s.response.error((f"{l}, error: "+output.replace("\n", ". ").replace(". .", ".")+".)").replace(". .",".").replace("\r","").replace("..","."))

			
		
		#
	def edit_comment(self, path=None, passphrase=None, comment=None):

		# check specific.
		if self.specific:
			if path == None: path = self.private_key

		# checks.
		path = Formats.denitialize(path)
		passphrase = Formats.denitialize(passphrase)
		comment = Formats.denitialize(comment)
		response = dev0s.response.parameters.check({
			"old":old,
			"passphrase":passphrase,
			"comment":comment,
		}, traceback=self.__traceback__(function="edit_comment"))
		if not response["success"]: return response

		# check dir.
		if os.path.isdir(path): path += '/private_key'
		
		# do.
		output = utils.__execute__(["ssh-keygen", "-c", "-P", passphrase, "-C", comment, "-f", path])

		# check fails.
		if "incorrect passphrase supplied" in output:
			return dev0s.response.error(f"Provided an incorrect passphrase for key [{path}].")
		elif "No such file or directory" in output:
			return dev0s.response.error(f"Key [{path}] does not exist.")
		
		# check success.	
		elif "Comment '" in output and "' applied" in output:
			return dev0s.response.success(f"Successfully edited the comment of key [{path}].")

		# unknown.
		else:
			l = f"Failed to edit the comment of key [{path}]"
			return dev0s.response.error((f"{l}, error: "+output.replace("\n", ". ").replace(". .", ".")+".)").replace(". .",".").replace("\r","").replace("..","."))

			
		
		#
	def generate(self, path=None, passphrase=None, comment=""):

		# check specific.
		if self.specific:
			if path == None: path = gfp.base(path=self.private_key)

		# checks.
		directory = Formats.denitialize(path)
		passphrase = Formats.denitialize(passphrase)
		comment = Formats.denitialize(comment)
		if directory[len(directory)-1] != "/": directory += "/"
		response = dev0s.response.parameters.check({
			"path":directory,
		}, traceback=self.__traceback__(function="generate"))
		if not response["success"]: return response

		# check arguments.
		if passphrase in [False, None, "", "null", "None", "none"]:
			passphrase = '""'

		# create dir.
		if directory != None and Files.exists(f"{directory}/public_key"): 
			return dev0s.response.error(f"Public key [{directory}/public_key] already exists.")
		elif directory != None and Files.exists(f"{directory}/private_key"): 
			return dev0s.response.error(f"Private key [{directory}/private_key] already exists.")
		elif directory != None and not Files.exists(directory): 
			os.mkdir(directory)
			Files.chmod(path=directory, permission=700, sudo=True)
			Files.chown(path=directory, owner=dev0s.defaults.vars.user, group=None, sudo=True)

		# options.
		private_key = f'{directory}/private_key'
		public_key = f'{directory}/public_key'
		identity_file = f'-f "{private_key}"'
		if comment == None: comment = ""
		if "[#id:" not in comment: comment += f" [#id:{String().generate(length=48, digits=True, capitalize=True, special=False)}]"
		comment = '-C "{}"'.format(comment)
		passphrase = f'-N "{passphrase}"'#utils.__string_to_bash__(passphrase)

		# execute.
		os.system(f'ssh-keygen -q -v -o -a 100 -t ed25519 {passphrase} {identity_file} {comment}')
		if not Files.exists(private_key): 
			return dev0s.response.error(f"Failed to generate key [{directory}].")

		# permissions.
		Files.chmod(path=private_key, permission=600, sudo=True)
		Files.chown(path=private_key, owner=dev0s.defaults.vars.user, group=None, sudo=True)
		os.system(f"mv '{private_key}.pub' '{public_key}'")
		if not Files.exists(public_key): 
			return dev0s.response.error(f"Failed to move private key [{private_key}].")
		Files.chmod(path=public_key, permission=640, sudo=True)
		Files.chown(path=public_key, owner=dev0s.defaults.vars.user, group=None, sudo=True)

		# response.
		return dev0s.response.success(f"Successfully generated key [{directory}].")

		#
	def check(self, username=None, public_keys=[], reversed=False):

		# check specific.
		if self.specific:
			if username == None: username = self.username

		# check if already present.
		if username == None: username = dev0s.defaults.vars.user
		username = Formats.denitialize(username)
		ssh_dir = FilePath(f"{dev0s.defaults.vars.homes}/{username}/.ssh/")
		auth_keys = FilePath(f"{dev0s.defaults.vars.homes}/{username}/.ssh/authorized_keys")
		output = self.__load_keys__(username)
		for key in public_keys:
			key = key.replace("\n", "")
			if key not in [""]:
				if not reversed and key not in output:
					ssh_dir.permission.set(permission=700, sudo=True, silent=True) # silent for when non existant.
					auth_keys.permission.set(permission=600, sudo=True, silent=True) # silent for when non existant.
					auth_keys.ownership.set(owner=username, sudo=True)
					ssh_dir.ownership.set(owner=username, sudo=True)
					return dev0s.response.error(f'Public key [{key}] is not activated.')
				if reversed and key in output:
					ssh_dir.permission.set(permission=700, sudo=True, silent=True) # silent for when non existant.
					auth_keys.permission.set(permission=600, sudo=True, silent=True) # silent for when non existant.
					auth_keys.ownership.set(owner=username, sudo=True)
					ssh_dir.ownership.set(owner=username, sudo=True)
					return dev0s.response.error(f'Public key [{key}] is activated.')

		# set correct permission.
		ssh_dir.permission.set(permission=700, sudo=True, silent=True) # silent for when non existant.
		auth_keys.permission.set(permission=600, sudo=True, silent=True) # silent for when non existant.
		auth_keys.ownership.set(owner=username, sudo=True)
		ssh_dir.ownership.set(owner=username, sudo=True)

		# success.
		if not reversed:
			return dev0s.response.success(f'Successfully confirmed that the specfied {len(public_keys)} public key(s) are activated.')
		else:
			return dev0s.response.success(f'Successfully confirmed that the specfied {len(public_keys)} public key(s) are not activated.')
	def enable(self, username=None, public_keys=[]):

		# check specific.
		if self.specific:
			if username == None: username = self.username
			if public_keys == []:
				public_keys = Array(Files.load(self.public_key).split("\n")).clean(remove_first=[" "], remove_last=[" "])

		# check if already present.
		if username == None: username = dev0s.defaults.vars.user
		output = self.__load_keys__(username)
		new_keys = []
		for key in public_keys:
			key = key.replace("\n", "")
			if key not in [""]:
				if key not in output:
					output.append(key)
		self.__save_keys__(username, output)

		# check if added.
		response = self.check(username, public_keys)
		if response["error"] != None: return response
	
		# success.
		return dev0s.response.success(f'Successfully enabled {len(public_keys)} public key(s).')

		#
	def disable(self, username=None, public_keys=[]):

		# check specific.
		if self.specific:
			if username == None: username = self.username
			if public_keys == []:
				public_keys = Array(Files.load(self.public_key).split("\n")).clean(remove_first=[" "], remove_last=[" "])

		# check if already present.
		if username == None: username = dev0s.defaults.vars.user
		output = self.__load_keys__(username)
		new_keys = []
		for key in output:
			key = key.replace("\n", "")
			if key not in [""]:
				if key not in public_keys: new_keys.append(key)
		self.__save_keys__(username, new_keys)

		# check if added.
		response = self.check(username, public_keys, reversed=True)
		if response["error"] != None: return response
	
		# success.
		return dev0s.response.success(f'Successfully disabled {len(public_keys)} public key(s).')

		#
	def __load_keys__(self, username):

		# make readable.
		if username == None: username = dev0s.defaults.vars.user
		sudo = dev0s.defaults.vars.user != username or True
		ssh_dir = FilePath(f"{dev0s.defaults.vars.homes}/{username}/.ssh/")
		auth_keys = FilePath(f"{dev0s.defaults.vars.homes}/{username}/.ssh/authorized_keys")

		# checks.
		if not ssh_dir.exists(sudo=sudo):
			ssh_dir.create(
				directory=True,
				permission=770,
				owner=username,
				group=None,
				sudo=sudo,)
		if not auth_keys.exists(sudo=sudo):
			auth_keys.create(
				directory=False,
				data="",
				permission=770,
				owner=username,
				group=None,
				sudo=sudo,)

		ssh_dir.permission.set(permission=770, sudo=sudo, silent=True) # silent for when non existant.
		auth_keys.permission.set(permission=770, sudo=sudo, silent=True) # silent for when non existant.
		auth_keys.ownership.set(owner=dev0s.defaults.vars.user, sudo=sudo)
		ssh_dir.ownership.set(owner=dev0s.defaults.vars.user, sudo=sudo)

		if sudo: command = ["sudo"]
		else: command = []
		output = utils.__execute__(command + ["cat", f"{dev0s.defaults.vars.homes}/{username}/.ssh/authorized_keys"], return_format="array")
		return output

		#
	def __save_keys__(self, username, public_keys):

		# make readable.
		if username == None: username = dev0s.defaults.vars.user
		sudo = dev0s.defaults.vars.user != username or True
		ssh_dir = FilePath(f"{dev0s.defaults.vars.homes}/{username}/.ssh/")
		auth_keys = FilePath(f"{dev0s.defaults.vars.homes}/{username}/.ssh/authorized_keys")

		# checks.
		if not ssh_dir.exists(sudo=sudo):
			ssh_dir.create(
				directory=True,
				permission=770,
				owner=username,
				group=None,
				sudo=sudo,)
		if not auth_keys.exists(sudo=sudo):
			auth_keys.create(
				directory=False,
				data="",
				permission=770,
				owner=username,
				group=None,
				sudo=sudo,)

		f = File(path="/tmp/file")
		new = []
		for public_key in public_keys:
			new.append(public_key.replace("\n",''))
		f.save(Array(path=False, array=new).string(joiner="\n"))
		os.system(f"sudo mv {f.file_path.path} {auth_keys.path}")

		ssh_dir.permission.set(permission=700, sudo=sudo, silent=True) # silent for when non existant.
		auth_keys.permission.set(permission=600, sudo=sudo, silent=True) # silent for when non existant.
		auth_keys.ownership.set(owner=dev0s.defaults.vars.user, sudo=sudo)
		ssh_dir.ownership.set(owner=dev0s.defaults.vars.user, sudo=sudo)

		#

# Initialized classes.
keys = Keys()

"""

# --------------------
# SSH Key.

# generate a key.
response = keys.generate(path="/path/to/mykey/", passphrase="passphrase123!", comment="my key")

# edit the passphrase of a key.
response = keys.edit_passphrase(path="/path/to/mykey/private_key", new="Passphrase123!", old="passphrase123!")

"""









#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
import os
from dev0s.shortcuts import *
from ssht00ls.classes.config import * 
from ssht00ls.classes.ssh import ssh
from ssht00ls.classes.ssync.index import index
from ssht00ls.classes.ssh.utils import execute

# settings.
INCLUDE = []# '.git', 'dist', "'*.egg-info'"
EXCLUDE = ['__pycache__', '.DS_Store']

# serialize path.
def serialize_path(path, append_last_slash=False):
	# keep in mind the file is saved by python and then executed.
	array = []
	for i in [" ","!","?","@","#","$","&","(",")","[","]","{","}"]:
		array.append([i, f"\\{i}"],)
	for x,y in array+[
		#[" ", "\\ "],
	]:
		path = path.replace(x,y)
	while True:
		if append_last_slash and len(path) > 0 and path[len(path)-1] != "/": path = path+"/"
		else: break
	return path

# get the size of a dir.
def size(path, alias=None, log_level=0):
	if alias == None:
		return dev0s.response.success(f"Successfully retrieved the size of {path}.", {
			"size":FilePath(path).size(mode="MB"),
		})
	else:
		return execute(
			command=f"""ssh {DEFAULT_SSH_OPTIONS} {alias} ' python3 /usr/local/lib/ssht00ls/lib/utils/size {path} --non-interactive ' '""",
			message=f"Successfully retrieved the size of [{alias}:{path}].",
			log_level=log_level,
			serialize=True,)

# pull.
def pull(
	# the local path.
	path=None, 
	# the ssht00ls alias.
	alias=None, 
	# the remote path.
	remote=None, 
	# exlude subpaths (list) (leave None to exclude none).
	exclude=[],
	# include subpaths (list) (leave None to include none).
	include=[],
	# path is directory boolean (leave None to parse automatically).
	directory=True,
	empty_directory=False,
	# update deleted files.
	delete=False,
	# forced mode.
	forced=False,
	# version control.
	safe=False,
	# accept new hosts keys.
	accept_new_host_keys=True,
	# checks.
	checks=True,
	# log level.
	log_level=dev0s.defaults.options.log_level,
	# get the command in str.
	command=False,
):	
	# checks.
	if exclude != None: exclude += EXCLUDE
	if include != None: include += INCLUDE
	if checks:

		# check alias.
		path = serialize_path(gfp.clean(path))
		remote = serialize_path(gfp.clean(remote))
		#response = aliases.check(alias)
		#if not response.success: return response
		
		# check encryption activated.
		#if not ssht00ls_agent.activated:
		#	return dev0s.response.error(f"The {ssht00ls_agent.id} encryption requires to be activated.")

		# check passphrase.
		#if CONFIG["aliases"][alias]["smartcard"] in [True, "true", "True"]:
		#	response = ssht00ls_agent.encryption.decrypt(CONFIG["aliases"][alias]["passphrase"])
		#else:
		#	response = ssht00ls_agent.encryption.decrypt(CONFIG["aliases"][alias]["passphrase"])
		#if not response.success: return response
		#passphrase = response.decrypted.decode()
		
		# tests.
		#response = agent.add(private_key=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
		#if not response["success"]: return response
		response = ssh.utils.test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response
		response = ssh.utils.test_path(alias=alias, path=remote, accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response

		# dir.
		if directory == None: 
			response = ssh.utils.test_dir(alias=alias, path=remote, accept_new_host_keys=accept_new_host_keys)
			if not response.success and "not a directory" not in response.error: return response
			elif response.success:
				directory = True
			else: directory = False
			tested = True
		elif directory:
			response = ssh.utils.test_dir(alias=alias, path=remote, accept_new_host_keys=accept_new_host_keys)
			if not response.success: return response
			tested = True

	# check base.
	base = FilePath(path).base(back=1)
	if not Files.exists(base): 
		os.system(f"mkdir -p {base}")
		if not Files.exists(base): 
			return dev0s.response.error(f"Failed to create pull base {base}.")
		if log_level >= 3:
			print(f"Created directory {base}.")

	# fix rsync timestamp bug.
	if empty_directory and directory and not Files.exists(path):
		os.system(f"mkdir -p {path}")

	# options.
	exclude_str = Array(array=exclude).string(joiner=" --exclude ", sum_first=True)
	include_str = Array(array=include).string(joiner=" --include ", sum_first=True)
	delete_str = Boolean(delete).string(true="--delete", false="")
	lremote = serialize_path(gfp.clean(remote), append_last_slash=directory)
	lpath = serialize_path(gfp.clean(path), append_last_slash=directory)
	_command_ = f"rsync -{Boolean(directory).string(true='a', false='')}zqt '{alias}:{lremote}' '{lpath}' {exclude_str} {include_str} {delete_str} --timeout={SSH_TIMEOUT}"
	#_command_ = f"rsync -azqtr '{alias}:{lremote}' '{lpath}' {exclude_str} {include_str} {delete_str}"

	# execute.
	if command: return _command_
	else:
		return execute(
			command=_command_,
			loader=f"Pulling [{alias}:{remote}] to [{path}]",
			message=f"Successfully pulled [{alias}:{remote}] to [{path}].",
			log_level=log_level,
		)

	#

# push.
def push(
	# the local path.
	path=None, 
	# the ssht00ls alias.
	alias=None, 
	# the remote path.
	remote=None, 
	# exlude subpaths (list) (leave None to exclude none).
	exclude=[],
	# include subpaths (list) (leave None to include none).
	include=[],
	# path is directory boolean (leave None to parse automatically).
	directory=True,
	empty_directory=False,
	# update deleted files.
	delete=False,
	# forced mode.
	forced=False,
	# version control.
	safe=False,
	# accept new hosts keys.
	accept_new_host_keys=True,
	# checks.
	checks=True,
	check_base=True,
	# log level.
	log_level=dev0s.defaults.options.log_level,
	# get the command in str.
	command=False,
):
	# checks.
	if exclude != None: exclude += EXCLUDE
	if include != None: include += INCLUDE
	if checks:

		# check alias.
		path = serialize_path(gfp.clean(path))
		remote = serialize_path(gfp.clean(remote))
		#response = aliases.check(alias)
		#if not response.success: return response
		
		# check encryption activated.
		#if not ssht00ls_agent.activated:
		#	return dev0s.response.error(f"The {ssht00ls_agent.id} encryption requires to be activated.")
		
		# check passphrase.
		#if CONFIG["aliases"][alias]["smartcard"] in [True, "true", "True"]:
		#	response = encrypion.encryption.decrypt(CONFIG["aliases"][alias]["passphrase"])
		#else:
		#	response = encrypion.encryption.decrypt(CONFIG["aliases"][alias]["passphrase"])
		#if not response.success: return response
		#passphrase = response.decrypted.decode()
		
		# tests.
		#response = agent.add(private_key=CONFIG["aliases"][alias]["private_key"], passphrase=passphrase)
		#if not response["success"]: return response
		response = ssh.utils.test_ssht00ls(alias=alias, accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response
		response = ssh.utils.test_path(alias=alias, path=FilePath(remote).base(), accept_new_host_keys=accept_new_host_keys)
		if not response.success: return response

		# dir.
		if directory == None: directory = os.path.isdir(path)
		elif directory and not os.path.isdir(path):
			return dev0s.response.error(f"Path {path} is not a directory.")

	# check remote base.
	# must be excluded from the checks == False.
	base = FilePath(remote).base(back=1)
	if check_base:
		response = ssh.utils.test_dir(alias=alias, path=base, accept_new_host_keys=accept_new_host_keys, create=True, checks=False)
		if not response.success: return response
		if response.created and log_level >= 3: print(f"Created remote directory {base}.")

	# options.
	exclude_str = Array(array=exclude).string(joiner=" --exclude ", sum_first=True)
	include_str = Array(array=include).string(joiner=" --include ", sum_first=True)
	delete_str = Boolean(delete).string(true="--delete", false="")
	lremote = serialize_path(gfp.clean(remote), append_last_slash=directory)
	lpath = serialize_path(gfp.clean(path), append_last_slash=directory)
	_command_ = f"rsync -{Boolean(directory).string(true='a', false='')}zqt '{lpath}' '{alias}:{lremote}' {exclude_str} {include_str} {delete_str} --timeout={SSH_TIMEOUT}"
	#_command_ = f"rsync -azqtr --rsh=ssh '{lpath}' '{alias}:{lremote}' {exclude_str} {include_str} {delete_str}"
	
	# execute.
	if command: return _command_
	else:
		return execute(
			command=_command_,
			loader=f"Pushing [{path}] to [{alias}:{remote}].",
			message=f"Successfully pushed [{path}] to [{alias}:{remote}].",
			log_level=log_level,
		)

	#

# main.
if __name__ == "__main__":
	a=1

#
	


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
from ssht00ls.classes.aliases import aliases
from ssht00ls.classes.ssh.tunnel import Tunnel
import ssht00ls.classes.ssh.utils as ssh_utils 

# the ssh object class.
class SSH(Traceback):
	def __init__(self,
		# initialize as specific not global (optional).
		# 	the alias.
		alias=None,
	):

		# defaults.
		Traceback.__init__(self, traceback="ssht00ls.ssh", raw_traceback="ssht00ls.classes.ssh.SSH")	

		# modules.
		self.utils = ssh_utils

		# objects.
		self.tunnel = Tunnel(
			alias=alias,)

		# specific args.
		self.specific = alias != None
		self.alias = alias

		#
	def session(self, 
		alias=None,
	):

		# check specific.
		if self.specific:
			if alias == None: alias = self.alias

		# checks.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="session"),
			parameters={
				"alias":alias,
			})
		if not response.success: return response

		# session.
		os.system(f"ssh {DEFAULT_SSH_OPTIONS} {alias}")

		# handler.
		return dev0s.response.success(f"Successfully started ssh session [{alias}].")

		#
	def command(self,
		# Alias:
		#   the alias.
		alias=None,
		# 
		# Command:
		#   the command in str.
		command="ls",
		#
		# Options:
		#   asynchronous process.
		async_=False,
		#	await asynchronous child (sync process always awaits).
		wait=False,
		#	kill process when finished (async that is not awaited is never killed).
		kill=True,
		#   the subprocess shell parameter.
		shell=False,
		#   serialize output to dict (expect literal dictionary / json output).
		serialize=False,
		# accept new host keys.
		accept_new_host_keys=True,
		#
		# Input (sync only):
		#   send input to the command.
		#	  undefined: send no input & automatically await the process since input is always sync.
		#	  dict instance: selects "and" mode ; send expected inputs and their value & return error when one of them is missing.
		#	  list[dict] instance: send all dictionaries in the list (default dict behaviour so one of the keys in each dict is expected).
		input=None,
		#   the input timeout (float) (list with floats by index from input)
		timeout=2.0,
		#   do not throw an error when the input is missing or not expected when optional is disabled (bool).
		optional=False, 
		#	apped default accept host keys input.
		append_default_input=True,
		#
		# Logging.
		# the success message (leave None to use the default).
		message=None,
		#   loader message.
		loader=None,
		#   the log level.
		log_level=dev0s.defaults.log_level(default=0),
		#
		# System functions.
		#   add additional attributes to the spawn object.
		__spawn_attributes__={},
	):
		
		# check specific.
		if self.specific:
			if alias == None: alias = self.alias

		# parse alias.
		response = aliases.info(alias=alias)
		if not response.success: return response
		info = response.info

		# add __spawn_attributes__.
		for key,value in {
			"alias":alias,
			"ip":info["ip"],
			"port":info["port"],
		}.items():
			try: __spawn_attributes__[str(key)]
			except KeyError: __spawn_attributes__[str(key)] = value

		# checks.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="command"),
			parameters={
				"alias":alias,
				"command":command,
			})
		if not response.success: return response

		# command.
		response = self.utils.execute(
			command=f"""ssh {DEFAULT_SSH_OPTIONS} {alias} ' {command} ' """,
			async_=async_,
			wait=wait,
			kill=kill,
			shell=shell,
			serialize=serialize,
			accept_new_host_keys=accept_new_host_keys,
			input=input,
			timeout=timeout,
			optional=optional,
			message=message,
			loader=loader,
			log_level=log_level,)

		# handler.
		if log_level >= 1:
			if response.success: print(response.output)
			else: print(response.error)
		return response
	
		
# initialized objects.
ssh = SSH()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
import os
os.environ["SSHT00LS_WEBSERVER_IMPORT"] = "True"
from ssht00ls.classes.config import *
os.environ["SSHT00LS_WEBSERVER_IMPORT"] = "False"

# the ssync object class.
class WebServer(dev0s.database.WebServer):
	def __init__(self):
		dev0s.database.WebServer.__init__(self, serialized={
			"id":"ssht00ls-agent",
			"path":f"{DATABASE}/.cache/",
			"host":"127.0.0.1",
			"default":{},
			"port":52379,
		})
	# could be expanded.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
from ssht00ls.classes import utils
from ssht00ls.classes.smartcards import smartcards

# the sshfs object class.
class SSHFS(Traceback):
	def __init__(self,
	):

		# defaults.
		Traceback.__init__(self, traceback="ssht00ls.sshfs", raw_traceback="ssht00ls.classes.sshfs.SSHFS")	

		#
	def mount(self, 
		# the directory paths.
		remote=None, 
		path=None, 
		# the ssh params.
		# option 1:
		alias=None,
		# option 2:
		username=None, 
		ip=None, 
		port=22,
		key_path=None,
	):

		# checks.
		base = ""
		if alias == None:
			response = dev0s.response.parameters.check(
				traceback=self.__traceback__(function="mount"), 
				parameters={
					"username":username,
					"ip":ip,
					"remote":remote,
					"path":path,
					"key_path":key_path,
					"port":port,
				})
			if not response["success"]: return response
			base += f"sshfs -p {port} -o IdentityFile={key_path} {username}@{ip}"
		else:
			response = dev0s.response.parameters.check(
				traceback=self.__traceback__(function="mount"), 
				parameters={
					"alias":alias,
					"remote":remote,
					"path":path,
				})
			if not response["success"]: return response
			base += f'sshfs {alias}'

		# do.
		command = f'{base}:{remote} {path}'
		print(f"COMMAND: [{command}]")
		output = utils.__execute_script__(command)
		#output = utils.__execute__(base + [f'{alias}:{remote}', path])
		#output = utils.__execute_script__(utils.__array_to_string__(base + [f'{alias}:{remote}', path], joiner="\n"))

		# check fails.
		if "mount_osxfuse: mount point " in output and "is itself" in output:
			return dev0s.response.error(f"Client path [{path}] is already mounted.")
		elif "No such file or directory" in output:
			return dev0s.response.error(f"Server path [{remote}] does not exist.")
		elif "" == output:
			if not Files.exists(path):
				return dev0s.response.error(f"Could not connect with server [{alias}].")
			# check success.	
			else:
				return dev0s.response.success(f"Successfully mounted directory [{path}].")

		# unknown.
		else:
			l = f"Failed to mount directory [{path}]"
			return dev0s.response.error((f"{l}, error: "+output.replace("\n", ". ").replace(". .", ".")+".)").replace(". .",".").replace("\r","").replace("..","."))
		
		#		
	def unmount(self, 
		# the client path.
		path=None, 
		# the forced umount option.
		forced=False, 
		# forced option may require sudo.
		sudo=False,
	):

		# checks.
		response = dev0s.response.parameters.check(
			traceback=self.__traceback__(function="unmount"),
			parameters={
				"path":path
			})
		if not response["success"]: return response
		command = []
		if sudo: command.append("sudo")
		command += ["umount"]
		if forced: command.append("-f")
		command += [path]
		output = utils.__execute__(command=command)
		if output != "":
			l = f"Failed to unmount directory [{path}]."
			return dev0s.response.error((f"{l}, error: "+output.replace("\n", ". ").replace(". .", ".")+".)").replace(". .",".").replace("\r","").replace("..","."))
		else:
			return dev0s.response.success(f"Successfully unmounted directory [{path}].")
		#
	
# Initialized classes.
sshfs = SSHFS()

"""

# --------------------
# SSHFS.
sshfs = SSHFS()

# mount a remote server directory.
response = sshfs.mount(
	# the directory paths.
	remote="/path/to/directory/", 
	path="/path/to/directory/", 
	# the ssh params.
	alias="administrator.myserver",)
	
# or without a created alias.
response = sshfs.mount(
	# the directory paths.
	remote="/path/to/directory/", 
	path="/path/to/directory/", 
	# the ssh params.
	username="administrator", 
	ip="0.0.0.0", 
	port=22,
	key_path="/path/to/mykey/private_key",)

# unmount a mounted directory.
response = sshfs.unmount(
	path="/path/to/directory/", 
	forced=False,
	sudo=False,)

"""









#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# import classes.
# keep this import style so you can also use the keys.Keys() etc directly.
from ssht00ls import classes
from ssht00ls.classes import ssht00ls_agent, network_info

# import initialized classes & objects.
from ssht00ls.classes.installation import installation
from ssht00ls.classes.sshd import sshd
from ssht00ls.classes.sshfs import sshfs
from ssht00ls.classes.scp import scp
from ssht00ls.classes.agent import agent
from ssht00ls.classes.keys import keys
from ssht00ls.classes.connections import connections
from ssht00ls.classes.smartcards import smartcards, SmartCard
from ssht00ls.classes.aliases import aliases
from ssht00ls.classes.ssh import ssh
from ssht00ls.classes.ssync import ssync
from ssht00ls.classes.smb import smb
from ssht00ls.classes.client import clients, Client

# shortcuts.
push = ssync.push
pull = ssync.pull

# source path & version.
from dev0s.shortcuts import Version, Directory, Files, gfp
source = Directory(gfp.base(__file__))
base = Directory(source.fp.base())
version = Version(Files.load(source.join(".version")))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import *
from ssht00ls.classes import utils
from ssht00ls.classes import agent, keys, aliases, ssh, smartcards, ssync, smb

# the client object class.
class Client(Traceback):
	def __init__(self,
		# the alias (required) (param #1).
		alias=None,
		# the username (optional if client already exists).
		username=None,
		# the public ip (optional if client already exists).
		public_ip=None,
		# the private ip (optional if client already exists).
		private_ip=None,
		# the public port (optional if client already exists).
		public_port=None,
		# the private port (optional if client already exists).
		private_port=None,
		# the path to the public key (optional if client already exists).
		public_key=None,
		# the path to the private key (optional if client already exists).
		private_key=None,
		# the smart card boolean (optional if client already exists).
		smartcard=False,
		# the smart card serial numbers (list) (optional if client already exists).
		serial_numbers=[],
		# pass parameters by dict.
		parameters={},
	):

		# defaults.
		Traceback.__init__(self, traceback="ssht00ls.Client", raw_traceback="ssht00ls.classes.client.Client")

		# parameters by dict.
		if parameters != {}:
			alias, username, public_ip, private_ip, public_port, private_port, public_key, private_key, smartcard, serial_numbers = Dictionary(parameters).unpack({
				"alias":None,
				"username":None,
				"public_ip":None,
				"private_ip":None,
				"public_port":None,
				"private_port":None,
				"public_key":None,
				"private_key":None,
				"smartcard":False,
				"serial_numbers":[],
			})

		# auto fill none & alias exists.
		response = aliases.aliases.info(alias)
		if response.success:
			if username == None: username = response.info["username"]
			if public_ip == None: public_ip = response.info["public_ip"]
			if private_ip == None: private_ip = response.info["private_ip"]
			if public_port == None: public_port = response.info["public_port"]
			if private_port == None: private_port = response.info["private_port"]
			if public_key == None: public_key = response.info["public_key"]
			if private_key == None: private_key = response.info["private_key"]
			if smartcard == None: smartcard = response.info["smartcard"]
			if serial_numbers != []: serial_numbers = response.info["serial_numbers"]

		# objects.
		self.alias = aliases.Aliases(
			alias=alias,
			username=username,
			public_ip=public_ip,
			private_ip=private_ip,
			public_port=public_port,
			private_port=private_port,
			private_key=private_key,
			public_key=public_key,
			smartcard=smartcard, 
			serial_numbers=serial_numbers, )
		self.agent = agent.Agent(
			private_key=private_key,
			public_key=public_key,
			smartcard=smartcard, )
		self.key = keys.Keys(
			private_key=private_key,
			public_key=public_key,
			username=username, )
		self.ssh = ssh.SSH(
			alias=alias, )
		self.ssync = ssync.SSync(
			alias=alias, )
		self.smartcards = smartcards.SmartCards()
		for serial_number in serial_numbers:
			self.smartcards[serial_number] = smartcards.SmartCard(serial_number=serial_number)
		self.smb = smb.SMB(
			alias=alias,)

		# shortcuts.
		self.pull = self.ssync.pull
		self.push = self.ssync.push

		# vars.
		try: 	self.public_key_data = Files.load(public_key)
		except: self.public_key_data = None
		try: 	self.public_key_id = self.public_key_data.split("[#id:")[1].split("]")[0]
		except: self.public_key_id = None

		#
	# functions.
	def generate(self, 
		# the new passphrase.
		passphrase=None,
		# the new smartcard pin.
		pin=None,
	):

		# generate key.
		response = self.key.generate(passphrase=passphrase)
		if not response.success: return response

		# create alias.
		response = self.create(
			passphrase=passphrase,
			pin=pin,)
		if not response.success: return response

		# handler.
		return dev0s.response.success(f"Successfully generated client {self.alias_}.")

		#
	def create(self, 
		# the new passphrase.
		passphrase=None,
		# the new smartcard pin.
		pin=None,
	):

		# create alias.
		response = self.alias.create(
			passphrase=passphrase,
			pin=pin,
			checks=False,)
		if not response.success: return response

		# check & add.
		response = self.check()
		if not response.success: return response		

		# handler.
		return dev0s.response.success(f"Successfully generated client {self.alias_}.")

		#
	def check(self):

		# check alias.
		response = self.alias.check(create=False, info={
			"username":self.username,
			"public_ip":self.public_ip,
			"private_ip":self.private_ip,
			"public_port":self.public_port,
			"private_port":self.private_port,
			"private_key":self.private_key,
			"public_key":self.public_key,
			"smartcard:bool,Boolean":self.smartcard,
		})
		if not response.success: return response

		# check added to agent.
		response = self.agent.check()
		if not response.success:
			if "is not added to the ssh agent" not in response.error: return response
			else:

				# get passphrase.
				passphrase, pin, no_passphrase = None, None, False
				response = self.alias.info()
				if not response.success: return response
				info = response.info
				if self.smartcard:
					if info["pin"] in [False, None, "", "none", "None"]:
						no_passphrase = True
					else:
						response = ssht00ls_agent.encryption.decrypt(info["pin"])
						if not response.success: return response
						pin = response.decrypted.decode()
				else:
					if info["passphrase"] in [False, None, "", "none", "None"]:
						no_passphrase = True
					else:
						response = ssht00ls_agent.encryption.decrypt(info["passphrase"])
						if not response.success: return response
						passphrase = response.decrypted.decode()

				# add agent key.
				if not no_passphrase:
					response = self.agent.add(
						passphrase=passphrase,
						pin=pin,)
					if not response.success: return response
				else:
					response = self.agent.add()
					if not response.success: return response

		# sync alias.
		response = self.alias.sync()
		if not response.success: return response

		# handler.
		return dev0s.response.success(f"Successfully checked client [{self.alias_}].")

		#
	def connect(self):
		return self.ssh.utils.test(alias=self.alias_)
	# properties.
	@property
	def exists(self):
		response = self.alias.check()
		if response.success: return True
		elif "does not exist." in response.error: return False
		else: raise ValueError(f"Unexpected error: {response.error}")
	@property
	def activated(self):
		response = self.agent.check()
		if response.success: return True
		elif "is not added to the ssh agent" in response.error: return False
		else: raise ValueError(f"Unexpected error: {response.error}")
	@property
	def id(self):
		return self.alias_
	@property
	def alias_(self):
		return self.alias.alias
	@property
	def username(self):
		return self.alias.username
	@property
	def public_ip(self):
		return self.alias.public_ip
	@property
	def private_ip(self):
		return self.alias.private_ip
	@property
	def public_port(self):
		return self.alias.public_port
	@property
	def private_port(self):
		return self.alias.private_port
	@property
	def ip(self):
		if self.public:
			return self.private_ip
		else:
			return self.private_ip
	@property
	def port(self):
		if self.public:
			return self.public_port
		else:
			return self.public_port
	@property
	def public(self):
		if None in [self.private_ip, self.public_ip]:
			return None
		try: return self.__public__
		except AttributeError: 
			self.__public__ = self.alias.public(public_ip=self.public_ip, private_ip=private_ip)
			return self.__public__
	@property
	def private_key(self):
		return self.alias.private_key
	@property
	def public_key(self):
		return self.alias.public_key
	@property
	def smartcard(self):
		return self.alias.smartcard
	@property
	def serial_numbers(self):
		return self.alias.serial_numbers

# the initialized clients.
class Clients(Traceback):
	def __init__(self,
		# select a preset of clients (append "*" to preset to use all).
		preset=["*"],
	):

		# defaults.
		Traceback.__init__(self, traceback="ssht00ls.clients", raw_traceback="ssht00ls.classes.client.Clients")

		# attributes.
		self.preset = preset
		self.__clients__ = {}

		#
	
	# initialize.
	def initialize(self, preset=None):
		if preset == None: preset = self.preset
		if preset in ["*","all"] or "*" in preset or  "all" in preset: 
			items = aliases.aliases.iterate()
		else:
			items = []
			for alias in preset:
				response = aliases.info(alias=alias)
				if not response.success: return response
				items.append([alias, response.info])
		self.__clients__ = {}
		for alias, info in items:
			self.__clients__[alias] = Client(parameters=info)
		return dev0s.response.success(f"Successfully initialized {len(self.__clients__)} client(s).")

		#

	# iterate.
	def __iter__(self):
		return iter(self.__clients__)
	def list(self):
		return list(self.__clients__.keys())
	def iterate(self, clients=[]):
		if clients == []:
			clients = list(self.__clients__.keys())
		items = []
		for i in clients:
			items.append([i, self.__clients__[i]])
		return items
		
		#

	# get & set client.
	def __setitem__(self, key, value):
		if isinstance(key, (int, Integer)):
			key = list(self.__clients__.keys())[key]
		self.__clients__[key] = value
	def __getitem__(self, key):
		if isinstance(key, (int, Integer)):
			key = list(self.__clients__.keys())[key]
		return self.__clients__[key]
		
		#

	# support [>=, <=, <, >] operators.
	def __gt__(self, clients):
		if isinstance(clients, int):
			a=1
		elif isinstance(clients, self.__class__):
			clients = len(clients.__clients__)
		elif not isinstance(clients, self.__class__):
			raise exceptions.FormatError(f"Can not compare object {self.__class__} & {clients.__class__}.")
		return len(self.__clients__) > clients
	def __ge__(self, clients):
		if isinstance(clients, int):
			a=1
		elif isinstance(clients, self.__class__):
			clients = len(clients.__clients__)
		elif not isinstance(clients, self.__class__):
			raise exceptions.FormatError(f"Can not compare object {self.__class__} & {clients.__class__}.")
		return len(self.__clients__) >= clients
	def __lt__(self, clients):
		if isinstance(clients, int):
			a=1
		elif isinstance(clients, self.__class__):
			clients = len(clients.__clients__)
		elif not isinstance(clients, self.__class__):
			raise exceptions.FormatError(f"Can not compare object {self.__class__} & {clients.__class__}.")
		return len(self.__clients__) < clients
	def __le__(self, clients):
		if isinstance(clients, int):
			a=1
		elif isinstance(clients, self.__class__):
			clients = len(clients.__clients__)
		elif not isinstance(clients, self.__class__):
			raise exceptions.FormatError(f"Can not compare object {self.__class__} & {clients.__class__}.")
		return len(self.__clients__) <= clients

		#

	# support [==, !=] operators.
	def __eq__(self, clients):
		if isinstance(clients, self.__class__):
			clients = clients.__clients__
		elif not isinstance(clients, self.__class__):
			return False
		return Dictionary(self.__clients__) == Dictionary(clients)
	def __ne__(self, clients):
		if isinstance(clients, self.__class__):
			clients = clients.__clients__
		elif not isinstance(clients, self.__class__):
			return True
		return Dictionary(self.__clients__) != Dictionary(clients)

		#

	# support +.
	def __concat__(self, clients):
		if isinstance(clients, self.__class__):
			clients = clients.__clients__
		elif not isinstance(value, self.__class__):
			raise exceptions.FormatError(f"Can not concat object {self.__class__} & {clients.__class__}.")
		return Dictionary(self.__clients__) + Dictionary(clients)

		#

	# support 'in' operator.
	def __contains__(self, alias):
		# always convert alias fo str client for aliases etc.
		if isinstance(alias, (list, Files.Array)):
			for i in alias:
				if str(i) in list(self.__clients__.keys()):
					return True
			return False
		else:
			return str(alias) in list(self.__clients__.keys())
		
		#
		
	# representations.
	def __repr__(self):
		return str(self)
	def __str__(self):
		return str(self.__clients__)
	def __int__(self):
		return int(self.__clients__)
	def __float__(self):
		return float(self.__clients__)
	def __bool__(self):
		return len(self.__clients__) > 0
	def __len__(self):
		return len(self.__clients__)

		#
	#

# initialized objects.
clients = Clients()
clients.initialize()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# imports.
from ssht00ls.classes.config import * 
import os, sys, requests, ast, json, pathlib, glob, string, getpass, django

# save config file safely.
def save_config_safely(backup=True, __loader__=None, __keyboard_interrupt__=None):
	if backup: save_config_backup_safely()		
	try:
		CONFIG.save()
	except KeyboardInterrupt as e:
		if __loader__ == None:
			__loader__ = dev0s.console.Loader("&RED&Do not interrupt!&END& Saving ssht00ls config file.")
		return save_config_safely(backup=False, __loader__=__loader__, __keyboard_interrupt__=e)
	if __loader__ != None: __loader__.stop()
	if __keyboard_interrupt__ != None:
		raise KeyboardInterrupt(__keyboard_interrupt__)

# save backup of config file safely.
def save_config_backup_safely(__loader__=None):
	path = DATABASE.join(".backups")
	if not Files.exists(path): Files.create(path, directory=True)
	path += "/config/"
	if not Files.exists(path): Files.create(path, directory=True)
	path += f"/{Date().date}"
	try:
		Files.save(path, CONFIG.dictionary, format="json")
	except KeyboardInterrupt as e:
		if __loader__ == None:
			__loader__ = dev0s.console.Loader("&RED&Do not interrupt!&END& Saving backup of ssht00ls config file.")
		return save_config_backup_safely(__loader__=__loader__)
	if __loader__ != None: __loader__.stop()
	fp = FilePath(gfp.base(path))
	if fp.size(format=int, mode="mb") >= 5:
		fp.delete(forced=True)
		fp.create(directory=True)

# check / start the ssh agent (due to circular import keep it over here for classes: [aliases]).
def ssh_agent():
	"""
	SSH_AUTH_SOCK = os.environ.get("SSH_AUTH_SOCK")
	SSH_AGENT_PID = os.environ.get("SSH_AGENT_PID")
	"""
	"""
	try:
		output = utils.__execute__([f"ssh-add", "-D"])
	except: a=1
	try:
		output = utils.__execute__([f"ssh-add", "-k"])
	except: a=1
	"""

	# version 2.
	try:
		output = utils.__execute__(f"ssh-agent")
		try: 
			SSH_AUTH_SOCK = output.split("SSH_AUTH_SOCK=")[1].split(";")[0]
			os.environ["SSH_AUTH_SOCK"] = SSH_AUTH_SOCK
		except: return None
		try: 
			SSH_AGENT_PID = output.split("SSH_AGENT_PID=")[1].split(";")[0]
			os.environ["SSH_AGENT_PID"] = SSH_AGENT_PID
		except: return None
	except: return None
	os.environ["SSH_AUTH_SOCK"] = SSH_AUTH_SOCK
	os.environ["SSH_AGENT_PID"] = SSH_AGENT_PID

# converting variables.
def __array_to_string__(array, joiner=" "):
	string = ""
	for i in array:
		if string == "": string = str(i)
		else: string += joiner+str(i)
	return string
def __string_to_boolean__(string):
	if string in ["true", "True", True]: return True
	elif string in ["false", "False", False]: return False
	else: raise ValueError(f"Could not convert string [{string}] to a boolean.")
def __string_to_bash__(string):
	a = string.replace('(','\(').replace(')','\)').replace("'","\'").replace(" ","\ ").replace("$","\$").replace("!","\!").replace("?","\?").replace("@","\@").replace("$","\$").replace("%","\%").replace("^","\^").replace("&","\&").replace("*","\*").replace("'","\'").replace('"','\"')       
	return a

# generation.
def __generate_pincode__(characters=6, charset=string.digits):
	return ''.join(random.choice(charset) for x in range(characters))
	#

# execute a shell command.
def __execute__(
	# the command in array.
	command=[],
	# wait till the command is pinished. 
	wait=False,
	# the commands timeout, [timeout] overwrites parameter [wait].
	timeout=None, 
	# the commands output return format: string / array.
	return_format="string", 
	# the subprocess.Popen.shell argument.
	shell=False,
	# pass a input string to the process.
	input=None,
):
	def __convert__(byte_array, return_format=return_format):
		if return_format == "string":
			lines = ""
			for line in byte_array:
				lines += line.decode()
			return lines
		elif return_format == "array":
			lines = []
			for line in byte_array:
				lines.append(line.decode().replace("\n","").replace("\\n",""))
			return lines

	# create process.
	if isinstance(command, str): command = command.split(' ')
	p = subprocess.Popen(
		command, 
		shell=shell,
		stdout=subprocess.PIPE,
		stderr=subprocess.PIPE,
		stdin=subprocess.PIPE,)
	
	# send input.
	if input != None:
		if isinstance(input, list):
			for s in input:
				p.stdin.write(f'{s}\n'.encode())
		elif isinstance(input, str):
			p.stdin.write(f'{input}\n'.encode())
		else: raise ValueError("Invalid format for parameter [input] required format: [string, array].")
		p.stdin.flush()
	
	# timeout.
	if timeout != None:
		time.sleep(timeout)
		p.terminate()
	
	# await.
	elif wait:
		p.wait()

	# get output.
	output = __convert__(p.stdout.readlines(), return_format=return_format)
	if return_format == "string" and output == "":
		output = __convert__(p.stderr.readlines(), return_format=return_format)
	elif return_format == "array" and output == []:
		output = __convert__(p.stderr.readlines(), return_format=return_format)
	return output

# execute a shell script.
def __execute_script__(
	# the script in string.
	script="",
	# wait till the command is pinished. 
	wait=False,
	# the commands timeout, [timeout] overwrites parameter [wait].
	timeout=None, 
	# the commands output return format: string / array.
	return_format="string", 
	# the subprocess.Popen.shell argument.
	shell=False,
	# pass a input string to the process.
	input=None,
):
	path = f"/tmp/shell_script.{__generate_pincode__(characters=32)}.sh"
	with open(str(path), "wb") as file:
		file.write(str(data))
	os.system(f"chmod +x {path}")
	output = __execute__(
		command=[f"sh", f"{path}"],
		wait=wait,
		timeout=timeout, 
		return_format=return_format, 
		shell=shell,
		input=input,)
	os.system(f"rm -fr {path}")
	return output
